#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>

#include <linux/mmc/host.h>
#include <linux/mmc/card.h>
#include <linux/mmc/mmc.h>
#include <linux/semaphore.h>

#define	PROC_NAME	"dt10sdif"
#define	DT10SDIF_CMD	11

/* ========== Parameter ========== */
#define		DT_VARIABLE_BIT			0x2	/* Variabe */
#define		DT_EVENTTRIGGER_MASK	0x8	/* EventTrigger */
/* ============================================ */

static struct mmc_host *dt10sd_host = NULL;
static struct semaphore dt10_sem;

void _TP_BusOut( unsigned int addr, unsigned int dat )
{
	unsigned char buff[4];
	struct mmc_command cmd;
	int	i;

	if( dt10sd_host == NULL)
	{
		printk( KERN_ERR "dt10sdif write fail host is null\n" );
		return;
	}
	buff[0] = (unsigned char)(dat >> 8);
	buff[1] = (unsigned char)(dat);
	buff[2] = (unsigned char)(addr >> 8);
	buff[3] = (unsigned char)(addr);
	down(&dt10_sem);
	mmc_claim_host(dt10sd_host);
	for( i = 0; i < (4>>2); i++ )
	{
		memset(&cmd, 0, sizeof(struct mmc_command));
		cmd.opcode = DT10SDIF_CMD;
		cmd.flags = MMC_RSP_NONE;
		cmd.arg = buff[0] << 24;
		cmd.arg += buff[1] << 16;
		cmd.arg += buff[2] << 8;
		cmd.arg += buff[3];
		mmc_wait_for_cmd(dt10sd_host, &cmd, 3);
	}
	mmc_release_host(dt10sd_host);
	up(&dt10_sem);
}
EXPORT_SYMBOL(_TP_BusOut);

void _TP_MemoryOutput( unsigned int addr, unsigned int data, void *value, unsigned int size)
{
	int	i;
	unsigned char *p = value;

	/* Memory Dump Start */
	_TP_BusOut(addr | DT_VARIABLE_BIT, data);
	/* Memory Dump */
	if( size >= 256 ) size = 256;
	if (value) {
		for( i = 0; i < size; i += 2 ){
			data = p[i] ;
			if( (i + 1) < size ){
				data |= p[i+1] << 8;
			}
			addr = DT_VARIABLE_BIT ;		// Variable Bit
			if( i == 0 ){
				addr |= size << 8 ;			// Size
			}
			_TP_BusOut( addr, data );
		}
	}
	else{
		_TP_BusOut(addr, data);
	}
}
EXPORT_SYMBOL(_TP_MemoryOutput);

void _TP_EventTrigger( unsigned int addr, unsigned int dat)
{
	_TP_BusOut(addr | DT_EVENTTRIGGER_MASK, dat << 4);
}
EXPORT_SYMBOL(_TP_EventTrigger);

void dt10_sd_host_get(struct mmc_host *host)
{
	dt10sd_host = host;
	sema_init(&dt10_sem, 1);
}

static int proc_write(struct file *filp, const char *buf, unsigned long len, void *data)
{
	unsigned char bus_out_data[5];
	struct mmc_command cmd;
	int	i;

	if( dt10sd_host == NULL )
	{
		printk( KERN_ERR "dt10sdif write fail host is null\n" );
		return -EFAULT;
	}
	if( copy_from_user(bus_out_data, buf, 5) )
	{
		printk( KERN_ERR "dt10sdif write fail data ng\n" );
		return -EFAULT;
	}
	down(&dt10_sem);
	mmc_claim_host(dt10sd_host);
	for( i = 0; i < (len>>2); i++ )
	{
		memset(&cmd, 0, sizeof(struct mmc_command));
		cmd.opcode = DT10SDIF_CMD;
		cmd.flags = MMC_RSP_NONE;
		cmd.arg = bus_out_data[0] << 24;
		cmd.arg += bus_out_data[1] << 16;
		cmd.arg += bus_out_data[2] << 8;
		cmd.arg += bus_out_data[3];
		mmc_wait_for_cmd(dt10sd_host, &cmd, 3);
	}
	mmc_release_host(dt10sd_host);
	up(&dt10_sem);
	return len;
}

static int proc_read(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	*eof = 1;
	return 0;
}

static int __init dt10_sdif_init(void)
{
	struct proc_dir_entry* entry;

	entry = create_proc_entry(PROC_NAME, 0666, NULL);
	if( entry )
	{
		entry->write_proc = proc_write;
		entry->read_proc = proc_read;
	}
	else
	{
		printk( KERN_ERR "proc entry fail\n" );
		return -EBUSY;
	}
	printk( KERN_INFO "dt10sdif load OK\n" );
	return 0;
}

static void __exit dt10_sdif_exit(void)
{
	remove_proc_entry(PROC_NAME, NULL);
	printk( KERN_INFO "dt10sdif unload\n" );
}

module_init( dt10_sdif_init );
module_exit( dt10_sdif_exit );

MODULE_DESCRIPTION("dt10sdif");
MODULE_LICENSE("GPL2");

